module RenoteDac

  module Mailer

    InvalidTemplate = Class.new(StandardError)

    BATCH_ID                = 0
    BATCH_TEMPLATE          = 1
    BATCH_EMAIL             = 2
    BATCH_PARAMS            = 3
    BATCH_ATTACHMENT_PARAMS = 4

    def self.enqueue(template, email_address, params = {}, attachment_params = [])
      meta = Mailer.get_template_meta(template)

      RenoteDac::Email.create!(
        priority: meta[:priority],
        template: template,
        address: email_address,
        params: params,
        attachment_params: attachment_params
      )
    end

    def self.get_batch(size)
      raise 'Max batch size is 500' if size > 500

      RenoteDac::Email
        .where(error: nil)
        .order(priority: :asc, id: :asc)
        .limit(size)
        .pluck(:id, :template, :address, :params, :attachment_params)
    end

    def self.get_template_meta(template)
      meta = RenoteDac.configuration.templates[template.to_sym]
      raise RenoteDac::Mailer::InvalidTemplate unless meta
      meta
    end

    def self.base_url
      if Rails.env.production?
        RenoteDac.configuration.prod_base_url
      else
        RenoteDac.configuration.dev_base_url
      end
    end

    class Client
      def initialize
        @client = Postmark::ApiClient.new(
          RenoteDac.configuration.postmark_api_key,
          http_read_timeout: 20,
          http_open_timeout: 10,
          secure: true
        )
      end

      def process_batch(size = 100)
        batch = Mailer.get_batch(size)
        return 0 if batch.empty?

        ids = []
        messages = []

        batch.each do |pending|
          template               = pending[BATCH_TEMPLATE]
          params                 = pending[BATCH_PARAMS]
          meta                   = Mailer.get_template_meta(template)
          file_attachment_params = pending[BATCH_ATTACHMENT_PARAMS]

          ids << pending[BATCH_ID]

          messages << {
            from: RenoteDac.configuration.from_address,
            to: pending[BATCH_EMAIL],
            template_id: meta[:template_id],
            template_model: params,
            tag: template,
            attachments: file_attachment_params
          }
        end

        if Rails.env.production?
          responses = @client.deliver_in_batches(messages)
          handle_resp(ids, responses)
        else
          messages.each do |message|
            puts "Sending message... #{message[:to]}"
          end
          # Dummy sending
          RenoteDac::Email.where(id: ids).update_all(error: 0)

          # Live sending
          #
          # if you must send messages for real in development to make sure the
          # templates look good, do it with care, because this uses the same
          # server as in production. Don't send 5,000 emails on accident or
          # Sidney will personally find you.
          #
          # responses = @client.deliver_in_batches(messages)
          # puts responses.inspect
          # handle_resp(ids, responses)
        end

        batch.size
      end

      private

      def handle_resp(ids, responses)
        raise 'Postmark response size mis-match' if ids.size != responses.size

        # http://developer.postmarkapp.com/developer-api-overview.html#error-codes

        ActiveRecord::Base.transaction do
          responses.each_with_index do |response, index|
            RenoteDac::Email.where(id: ids[index]).update_all(error: response[:error_code], sent_at: Time.now.utc)
          end
        end
      end
    end
  end
end